@@ -1,26 +1,29 @@
 /*
  *  ser2net - A program for allowing telnet connection to serial ports
- *  Copyright (C) 2001-2016  Corey Minyard <minyard@acm.org>
+ *  Copyright (C) 2016-2020  Corey Minyard <minyard@acm.org>
  *  Copyright (C) 2016 Michael Heimpold <mhei@heimpold.de>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  SPDX-License-Identifier: GPL-2.0-only
  *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ *  In addition, as a special exception, the copyright holders of
+ *  ser2net give you permission to combine ser2net with free software
+ *  programs or libraries that are released under the GNU LGPL and
+ *  with code included in the standard release of OpenSSL under the
+ *  OpenSSL license (or modified versions of such code, with unchanged
+ *  license). You may copy and distribute such a system following the
+ *  terms of the GNU GPL for ser2net and the licenses of the other code
+ *  concerned, provided that you include the source code of that
+ *  other code when and as the GNU GPL requires distribution of source
+ *  code.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Note that people who make modified versions of ser2net are not
+ *  obligated to grant this special exception for their modified
+ *  versions; it is their choice whether to do so. The GNU General
+ *  Public License gives permission to release a modified version
+ *  without this exception; this exception also makes it possible to
+ *  release a modified version which carries forward this exception.
  */
 
-/* This file holds common LED code required to read the configuration file and
-   to dispatch calls from dataxfer to the actual LED driver. */
-
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
@@ -68,69 +71,6 @@ led_driver_register(struct led_driver_s *led_driver)
     return 0;
 }
 
-void
-handle_led(const char *name, char *cfg, int lineno)
-{
-    struct led_driver_s *driver;
-    struct led_s *new_led;
-    char *delim;
-
-    delim = strchr(cfg, ':');
-    if (!delim) {
-	syslog(LOG_ERR, "Couldn't parse LED definition for '%s' on %d",
-	       name, lineno);
-	return;
-    }
-    *delim++ = '\0';
-
-    driver = led_driver_by_name(cfg);
-    if (!driver) {
-	syslog(LOG_ERR, "Unknown LED driver '%s' for LED '%s' on %d",
-	       cfg, name, lineno);
-	return;
-    }
-
-    new_led = calloc(1, sizeof(*new_led));
-    if (!new_led) {
-	syslog(LOG_ERR, "Out of memory handling LED '%s' on %d", name, lineno);
-	return;
-    }
-
-    new_led->name = strdup(name);
-    if (!new_led->name) {
-	syslog(LOG_ERR, "Out of memory handling LED '%s' on %d", name, lineno);
-	free(new_led);
-	return;
-    }
-
-    new_led->driver = driver;
-
-    if (new_led->driver->init(new_led, delim, lineno) < 0) {
-	/* errors should be reported by driver itself */
-	free(new_led->name);
-	free(new_led);
-	return;
-    }
-
-    if (new_led->driver->configure) {
-	if (new_led->driver->configure(new_led->drv_data) < 0) {
-	    /*
-	     * errors should be reported by driver itself; however, we
-	     * cleanup here
-	     */
-	    if (new_led->driver->free)
-		new_led->driver->free(new_led);
-
-	    free(new_led->name);
-	    free(new_led);
-	    return;
-	}
-    }
-
-    new_led->next = leds;
-    leds = new_led;
-}
-
 struct led_s *
 find_led(const char *name)
 {
@@ -142,10 +82,71 @@ find_led(const char *name)
 	led = led->next;
     }
 
-    syslog(LOG_ERR, "LED '%s' not found, it will be ignored", name);
     return NULL;
 }
 
+int
+add_led(const char *name, const char *driverstr, const char * const *options,
+	int lineno)
+{
+    struct led_driver_s *driver;
+    struct led_s *led;
+
+    led = find_led(name);
+    if (led) {
+	syslog(LOG_ERR, "LED %s already exists on line %d\n", name, lineno);
+	return -1;
+    }
+
+    driver = led_driver_by_name(driverstr);
+    if (!driver) {
+	syslog(LOG_ERR, "Unknown LED driver '%s' for LED '%s' on %d",
+	       driverstr, name, lineno);
+	return -1;
+    }
+
+    led = calloc(1, sizeof(*led));
+    if (!led) {
+	syslog(LOG_ERR, "Out of memory handling LED '%s' on %d", name, lineno);
+	return -1;
+    }
+
+    led->name = strdup(name);
+    if (!led->name) {
+	syslog(LOG_ERR, "Out of memory handling LED '%s' on %d", name, lineno);
+	free(led);
+	return -1;
+    }
+
+    led->driver = driver;
+
+    if (led->driver->init(led, options, lineno) < 0) {
+	/* errors should be reported by driver itself */
+	free(led->name);
+	free(led);
+	return -1;
+    }
+
+    if (led->driver->configure) {
+	if (led->driver->configure(led->drv_data, lineno) < 0) {
+	    /*
+	     * errors should be reported by driver itself; however, we
+	     * cleanup here
+	     */
+	    if (led->driver->free)
+		led->driver->free(led);
+
+	    free(led->name);
+	    free(led);
+	    return -1;
+	}
+    }
+
+    led->next = leds;
+    leds = led;
+    return 0;
+}
+
 void
 free_leds(void)
 {
