@@ -1,20 +1,26 @@
 /*
  *  ser2net - A program for allowing telnet connection to serial ports
- *  Copyright (C) 2001  Corey Minyard <minyard@acm.org>
+ *  Copyright (C) 2001-2020  Corey Minyard <minyard@acm.org>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  SPDX-License-Identifier: GPL-2.0-only
  *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ *  In addition, as a special exception, the copyright holders of
+ *  ser2net give you permission to combine ser2net with free software
+ *  programs or libraries that are released under the GNU LGPL and
+ *  with code included in the standard release of OpenSSL under the
+ *  OpenSSL license (or modified versions of such code, with unchanged
+ *  license). You may copy and distribute such a system following the
+ *  terms of the GNU GPL for ser2net and the licenses of the other code
+ *  concerned, provided that you include the source code of that
+ *  other code when and as the GNU GPL requires distribution of source
+ *  code.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Note that people who make modified versions of ser2net are not
+ *  obligated to grant this special exception for their modified
+ *  versions; it is their choice whether to do so. The GNU General
+ *  Public License gives permission to release a modified version
+ *  without this exception; this exception also makes it possible to
+ *  release a modified version which carries forward this exception.
  */
 
 /* This file holds the code that reads the configuration file and
@@ -29,23 +35,15 @@
 #include <syslog.h>
 #include <stdarg.h>
 #include <limits.h>
+#include <errno.h>
+#include <gensio/gensio.h>
+#include <gensio/argvutils.h>
 
+#include "ser2net.h"
 #include "dataxfer.h"
 #include "readconfig.h"
-#include "utils.h"
-#include "telnet.h"
 #include "led.h"
-
-#ifdef HAVE_OPENIPMI
-#include <OpenIPMI/ipmi_conn.h>
-#include <OpenIPMI/ipmi_sol.h>
-#endif
-
-#define PORT_BUFSIZE	64	/* Default data transfer buffer size */
-
-extern char *config_port;
-
-static int config_num = 0;
+#include "defaults.h"
 
 static int lineno = 0;
 
@@ -251,15 +249,6 @@ struct tracefile_s
     struct tracefile_s *next;
 };
 
-#if HAVE_DECL_TIOCSRS485
-struct rs485conf_s
-{
-    char *name;
-    struct serial_rs485 conf;
-    struct rs485conf_s *next;
-};
-#endif
-
 /* All the tracefiles in the system. */
 struct tracefile_s *tracefiles = NULL;
 
@@ -320,15 +309,20 @@ free_tracefiles(void)
     }
 }
 
-#if HAVE_DECL_TIOCSRS485
+struct rs485conf
+{
+    char *name;
+    char *str;
+    struct rs485conf *next;
+};
+
 /* All the RS485 configs in the system. */
-struct rs485conf_s *rs485confs = NULL;
+struct rs485conf *rs485confs = NULL;
 
 static void
 handle_rs485conf(char *name, char *str)
 {
-    struct rs485conf_s *new_rs485conf;
-    uint8_t rts_on_send, rx_during_tx;
+    struct rs485conf *new_rs485conf;
 
     new_rs485conf = malloc(sizeof(*new_rs485conf));
     if (!new_rs485conf) {
@@ -343,59 +337,36 @@ handle_rs485conf(char *name, char *str)
 	goto out_err;
     }
 
-    if (sscanf(str, "%10u:%10u:%1hhu:%1hhu",
-               &new_rs485conf->conf.delay_rts_before_send,
-               &new_rs485conf->conf.delay_rts_after_send,
-               &rts_on_send,
-               &rx_during_tx) != 4) {
-	syslog(LOG_ERR, "Couldn't parse RS485 config on %d", lineno);
+    new_rs485conf->str = strdup(str);
+    if (!new_rs485conf->str) {
+	syslog(LOG_ERR, "Out of memory handling rs485 config on %d", lineno);
 	goto out_err;
     }
 
-    /* check, if flags have values 0 or 1 */
-    if (rts_on_send > 1) {
-	syslog(LOG_ERR, "RTS_ON_SEND parameter can be 0 or 1 on %d", lineno);
-	goto out_err;
-    }
-
-    if (rx_during_tx > 1) {
-	syslog(LOG_ERR, "RX_DURING_TX parameter can be 0 or 1 on %d", lineno);
-	goto out_err;
-    }
-
-    new_rs485conf->conf.flags = SER_RS485_ENABLED;
-
-    if (rts_on_send) {
-        new_rs485conf->conf.flags |= SER_RS485_RTS_ON_SEND;
-    } else {
-        new_rs485conf->conf.flags |= SER_RS485_RTS_AFTER_SEND;
-    }
-
-    if (rx_during_tx) {
-        new_rs485conf->conf.flags |= SER_RS485_RX_DURING_TX;
-    }
-
     new_rs485conf->next = rs485confs;
     rs485confs = new_rs485conf;
     return;
 
  out_err:
+    if (new_rs485conf->str)
+	free(new_rs485conf->str);
     if (new_rs485conf->name)
 	free(new_rs485conf->name);
     free(new_rs485conf);
 }
 
-struct serial_rs485 *
+char *
 find_rs485conf(const char *name)
 {
-    struct rs485conf_s *new_rs485conf = rs485confs;
+    struct rs485conf *rs485 = rs485confs;
 
-    while (new_rs485conf) {
-        if (strcmp(name, new_rs485conf->name) == 0)
-            return &new_rs485conf->conf;
-        new_rs485conf = new_rs485conf->next;
+    while (rs485) {
+        if (strcmp(name, rs485->name) == 0)
+            return strdup(rs485->str);
+        rs485 = rs485->next;
     }
-    syslog(LOG_ERR, "RS485 configuration %s not found, it will be ignored", name);
+    syslog(LOG_ERR, "RS485 configuration %s not found, it will be ignored",
+	   name);
     return NULL;
 }
 
@@ -403,16 +374,14 @@ void
 free_rs485confs(void)
 {
     while (rs485confs) {
-	struct rs485conf_s *rs485conf = rs485confs;
+	struct rs485conf *rs485 = rs485confs;
 
-        rs485confs = rs485confs->next;
-        free(rs485conf->name);
-        free(rs485conf);
+        rs485 = rs485->next;
+        free(rs485->str);
+        free(rs485->name);
+        free(rs485);
     }
 }
-#else
-void free_rs485confs(void) { }
-#endif
 
 static int
 startswith(char *str, const char *test, char **strtok_data)
@@ -444,292 +413,9 @@ static struct absout syslog_eout = {
     .data = &lineno
 };
 
-struct enum_val
-{
-    char *str;
-    int val;
-};
-
-static int
-lookup_enum(struct enum_val *enums, const char *str, int len)
-{
-    while (enums->str != NULL) {
-	if (len == -1 && strcmp(enums->str, str) == 0)
-	    return enums->val;
-	if (strlen(enums->str) == len && strncmp(enums->str, str, len) == 0)
-	    return enums->val;
-	enums++;
-    }
-    return -1;
-}
-
-static struct enum_val speed_enums[] = {
-    { "300",	300 },
-    { "600",	600 },
-    { "1200",	1200 },
-    { "2400",	2400 },
-    { "4800",	4800 },
-    { "9600",	9600 },
-    { "19200",	19200 },
-    { "38400",	38400 },
-    { "57600",	57600 },
-    { "115200",	115200 },
-    { "230400",	230400 },
-    { "460800",	460800 },
-    { "500000",	500000 },
-    { "576000",	576000 },
-    { "921600",	921600 },
-    { "1000000",1000000 },
-    { "1152000",1152000 },
-    { "1500000",1500000 },
-    { "2000000",2000000 },
-    { "2500000",2500000 },
-    { "3000000",3000000 },
-    { "3500000",3500000 },
-    { "4000000",4000000 },
-    { NULL },
-};
-
-int
-speedstr_to_speed(const char *speed)
-{
-    return lookup_enum(speed_enums, speed, -1);
-}
-
-struct enum_val parity_enums[] = {
-    { "NONE", PARITY_NONE },
-    { "EVEN", PARITY_EVEN },
-    { "ODD", PARITY_ODD },
-    { "none", PARITY_NONE },
-    { "even", PARITY_EVEN },
-    { "odd", PARITY_ODD },
-    { "MARK", PARITY_MARK },
-    { "SPACE", PARITY_SPACE },
-    { "mark", PARITY_MARK },
-    { "space", PARITY_SPACE },
-    { NULL }
-};
-
-#ifdef HAVE_OPENIPMI
-struct enum_val shared_serial_alert_enums[] = {
-    { "fail",		ipmi_sol_serial_alerts_fail },
-    { "deferred", 	ipmi_sol_serial_alerts_deferred },
-    { "succeed", 	ipmi_sol_serial_alerts_succeed },
-    { NULL }
-};
-#endif
-
-enum parity_vals
-lookup_parity(const char *str)
-{
-    return lookup_enum(parity_enums, str, -1);
-}
-
-enum default_type { DEFAULT_INT, DEFAULT_BOOL, DEFAULT_ENUM, DEFAULT_STR };
-
-struct default_data
-{
-    const char *name;
-    enum default_type type;
-    int min;
-    int max;
-    struct enum_val *enums;
-    union {
-	int intval;
-	char *strval;
-    } val;
-    union {
-	int intval;
-	const char *strval;
-    } def;
-    const char *altname;
-};
-
-struct default_data defaults[] = {
-    /* serial device only */
-    { "stopbits",	DEFAULT_INT,	.min = 1, .max = 2, .def.intval = 1 },
-    { "databits",	DEFAULT_INT,	.min = 5, .max = 8, .def.intval = 8 },
-    { "parity",		DEFAULT_ENUM,	.enums = parity_enums,
-					.def.intval = PARITY_NONE },
-    { "xonxoff",	DEFAULT_BOOL,	.def.intval = 0 },
-    { "rtscts",		DEFAULT_BOOL,	.def.intval = 0 },
-    { "local",		DEFAULT_BOOL,	.def.intval = 0 },
-    { "hangup_when_done", DEFAULT_BOOL,	.def.intval = 0 },
-    /* Serial port and SOL */
-    { "speed",		DEFAULT_ENUM,	.enums = speed_enums,
-					.def.intval = 9600 },
-    { "nobreak",	DEFAULT_BOOL,	.def.intval = 0 },
-    /* All port types */
-    { "remctl",		DEFAULT_BOOL,	.def.intval = 0 },
-    { "telnet_brk_on_sync",DEFAULT_BOOL,.def.intval = 0 },
-    { "kickolduser",	DEFAULT_BOOL,	.def.intval = 0 },
-    { "chardelay",	DEFAULT_BOOL,	.def.intval = 1 },
-    { "chardelay-scale",DEFAULT_INT,	.min = 1, .max = 1000,
-					.def.intval = 20 },
-    { "chardelay-min",	DEFAULT_INT,	.min = 1, .max = 100000,
-					.def.intval = 1000 },
-    { "chardelay-max",	DEFAULT_INT,	.min = 1, .max = 1000000,
-					.def.intval = 20000 },
-    { "dev-to-net-bufsize", DEFAULT_INT,.min = 1, .max = 65536,
-					.def.intval = PORT_BUFSIZE,
-					.altname = "dev-to-tcp-bufsize" },
-    { "net-to-dev-bufsize", DEFAULT_INT,.min = 1, .max = 65536,
-					.def.intval = PORT_BUFSIZE,
-					.altname = "tcp-to-dev-bufsize" },
-    { "max-connections", DEFAULT_INT,	.min=1, .max=65536,
-					.def.intval = 1 },
-#ifdef HAVE_OPENIPMI
-    /* SOL only */
-    { "authenticated",	DEFAULT_BOOL,	.def.intval = 1 },
-    { "encrypted",	DEFAULT_BOOL,	.def.intval = 1 },
-    { "ack-timeout",	DEFAULT_INT,	.min = 1, .max = INT_MAX,
-					.def.intval = 1000000 },
-    { "ack-retries",	DEFAULT_INT,	.min = 1, .max = INT_MAX,
-					.def.intval = 10 },
-    { "shared-serial-alert", DEFAULT_ENUM, .enums = shared_serial_alert_enums,
-				   .def.intval = ipmi_sol_serial_alerts_fail },
-    { "deassert_CTS_DCD_DSR_on_connect", DEFAULT_BOOL, .def.intval = 0 },
-#endif
-    { "remaddr",	DEFAULT_STR,	.def.strval = "" },
-    { NULL }
-};
-
-
-static void
-setup_defaults(void)
-{
-    int i;
-
-    for (i = 0; defaults[i].name; i++) {
-	if (defaults[i].type == DEFAULT_STR) {
-	    if (defaults[i].val.strval) {
-		free(defaults[i].val.strval);
-		defaults[i].val.strval = NULL;
-	    }
-	} else {
-	    defaults[i].val.intval = defaults[i].def.intval;
-	}
-    }
-}
-
-static int cmp_default_name(struct default_data *def, const char *name)
-{
-    return (strcmp(def->name, name) == 0 ||
-	    (def->altname && strcmp(def->altname, name) == 0));
-}
-
-int
-find_default_int(const char *name)
-{
-    int i;
-
-    for (i = 0; defaults[i].name; i++) {
-	if (cmp_default_name(&defaults[i], name) &&
-			defaults[i].type != DEFAULT_STR)
-	    return defaults[i].val.intval;
-    }
-    abort();
-}
-
-char *
-find_default_str(const char *name)
-{
-    int i;
-
-    for (i = 0; defaults[i].name; i++) {
-	if (cmp_default_name(&defaults[i], name) &&
-			defaults[i].type == DEFAULT_STR) {
-	    const char *s = defaults[i].val.strval;
-	    if (!s)
-		s = defaults[i].def.strval;
-	    return strdup(s);
-	}
-    }
-    abort();
-}
-
-static void
-handle_new_default(const char *name, const char *str)
-{
-    int i, val, len;
-    char *end, *sval;
-    const char *s;
-
-    while (isspace(*str))
-	str++;
-    s = str;
-    while (!isspace(*s) && *s != '\0')
-	s++;
-    if (s == str) {
-	syslog(LOG_ERR, "No default value on %d", lineno);
-	return;
-    }
-    len = s - str;
-
-    for (i = 0; defaults[i].name; i++) {
-	if (!cmp_default_name(&defaults[i], name))
-	    continue;
-
-	switch (defaults[i].type) {
-	case DEFAULT_INT:
-	    val = strtoul(str, &end, 10);
-	    if (end != s) {
-		syslog(LOG_ERR, "Invalid integer value on %d", lineno);
-		return;
-	    }
-	    if (val < defaults[i].min || val > defaults[i].max) {
-		syslog(LOG_ERR, "Integer value out of range on %d, "
-		       "min is %d, max is %d",
-		       lineno, defaults[i].min, defaults[i].max);
-		return;
-	    }
-	    defaults[i].val.intval = val;
-	    break;
-
-	case DEFAULT_BOOL:
-	    val = strtoul(str, &end, 10);
-	    if (end == s)
-		defaults[i].val.intval = !!val;
-	    else if (len == 4 && ((strncmp(str, "true", 4) == 0) ||
-				  (strncmp(str, "TRUE", 4) == 0)))
-		defaults[i].val.intval = 1;
-	    else if (len == 5 && ((strncmp(str, "false", 5) == 0) ||
-				  (strncmp(str, "FALSE", 5) == 0)))
-		defaults[i].val.intval = 0;
-	    else
-		syslog(LOG_ERR, "Invalid integer value on %d", lineno);
-	    break;
-
-	case DEFAULT_ENUM:
-	    val = lookup_enum(defaults[i].enums, str, len);
-	    if (val == -1) {
-		syslog(LOG_ERR, "Invalid enumeration value on %d", lineno);
-		return;
-	    }
-	    defaults[i].val.intval = val;
-	    break;
-
-	case DEFAULT_STR:
-	    sval = strdup(str);
-	    if (!sval) {
-		syslog(LOG_ERR, "Out of memory processing default string on"
-		       " line %d", lineno);
-		return;
-	    }
-	    if (defaults[i].val.strval)
-		free(defaults[i].val.strval);
-	    defaults[i].val.strval = sval;
-	    break;
-	}
-	return;
-    }
-
-    syslog(LOG_ERR, "unknown default name '%s' on %d", name, lineno);
-}
-
 /*
  * This rather complicated variable is used to scan the string for
- * ":off", ":telnet", ":raw:", or ":rawlp:".  It's a basic state
+ * ":off", ":on", ":telnet", ":raw:", or ":rawlp:".  It's a basic state
  * machine where if a character in the first string "c" matches the
  * current character, you go to the state machine index in the
  * corresponding location giving by the character in string "next".
@@ -756,7 +442,7 @@ static struct {
     { "w",   "\x09" },		/* 0x09 */
     { ":l",  "\x00\x0a" },	/* 0x09 */
     { "p",   "\x0d" },		/* 0x0a */
-    { "f",   "\x0c" },		/* 0x0b */
+    { "fn",  "\x0c\x0d" },	/* 0x0b */
     { "f",   "\x0d" },		/* 0x0c */
     { ":",   "\x00" }		/* 0x0d */
 };
@@ -791,8 +477,11 @@ scan_for_state(char *str)
 int
 handle_config_line(char *inbuf, int len)
 {
-    char *portnum, *state, *timeout, *devname, *devcfg;
+    char *portnum, *state, *timeoutstr, *devname, *devcfg;
+    unsigned int timeout;
     char *strtok_data = NULL;
+    const char **devcfg_argv;
+    int err;
 
     if (len == 0)
 	/* Ignore empty lines */
@@ -887,21 +576,12 @@ handle_config_line(char *inbuf, int len)
     }
 
     if (startswith(inbuf, "CONTROLPORT", &strtok_data)) {
-	if (config_port)
-	    /*
-	     * The control port has already been configured either on the
-	     * command line or on a previous statement.  Only take the first.
-	     */
-	    goto out;
-	config_port = strdup(strtok_r(NULL, "\n", &strtok_data));
-	if (!config_port) {
-	    syslog(LOG_ERR, "Could not allocate memory for CONTROLPORT");
-	    goto out;
-	}
+	char *config_port = strtok_r(NULL, "\n", &strtok_data);
+
+	controller_init(config_port, NULL, NULL);
 	goto out;
     }
 
-#if HAVE_DECL_TIOCSRS485
     if (startswith(inbuf, "RS485CONF", &strtok_data)) {
         char *name = strtok_r(NULL, ":", &strtok_data);
         char *str = strtok_r(NULL, "\n", &strtok_data);
@@ -916,42 +596,102 @@ handle_config_line(char *inbuf, int len)
         handle_rs485conf(name, str);
         goto out;
     }
-#endif
 
     if (startswith(inbuf, "DEFAULT", &strtok_data)) {
 	char *name = strtok_r(NULL, ":", &strtok_data);
-	char *str = strtok_r(NULL, "\n", &strtok_data);
+	char *str = strtok_r(NULL, ":\n", &strtok_data);
+	char *class = strtok_r(NULL, "\n", &strtok_data);
+
 	if (name == NULL) {
 	    syslog(LOG_ERR, "No default name given on line %d", lineno);
 	    goto out;
 	}
-	handle_new_default(name, str);
+	if (str && strlen(str) == 0) {
+	    str = NULL;
+	}
+	if (class) {
+	    /* Watch out for trailing spaces. */
+	    unsigned int len = strlen(class);
+
+	    while(len > 0 && isspace(class[len]))
+		len--;
+	    class[len] = '\0';
+	}
+
+	err = gensio_set_default(so, class, name, str, 0);
+	if (err)
+	    syslog(LOG_ERR, "error setting default value on line %d for %s: %s",
+		   lineno, name, strerror(err));
+	goto out;
+    }
+
+    if (startswith(inbuf, "DELDEFAULT", &strtok_data)) {
+	char *name = strtok_r(NULL, ":", &strtok_data);
+	char *class = strtok_r(NULL, "\n", &strtok_data);
+
+	if (!name) {
+	    syslog(LOG_ERR, "No default name given on line %d", lineno);
+	    goto out;
+	}
+	if (!class || strcmp(class, "ser2net") == 0 ||
+			strcmp(class, "default") == 0) {
+	    syslog(LOG_ERR, "Can only delete class default on line %d", lineno);
+	    goto out;
+	}
+
+	err = gensio_del_default(so, class, name, false);
+	if (err)
+	    syslog(LOG_ERR, "error deleting default value on line %d for %s: %s",
+		   lineno, name, strerror(err));
 	goto out;
     }
 
     if (startswith(inbuf, "ROTATOR", &strtok_data)) {
 	char *name = strtok_r(NULL, ":", &strtok_data);
 	char *str = strtok_r(NULL, "\n", &strtok_data);
+	int portc;
+	const char **portv;
+
 	if (name == NULL) {
 	    syslog(LOG_ERR, "No rotator name given on line %d", lineno);
 	    goto out;
 	}
-	add_rotator(name, str, lineno);
+
+	err = gensio_str_to_argv(so, str, &portc, &portv, NULL);
+	if (err) {
+	    syslog(LOG_ERR, "Unable to allocate rotator argv on line %d",
+		   lineno);
+	    goto out;
+	}
+	err = add_rotator(&syslog_eout, name, name, portc, portv, NULL, lineno);
+	if (err)
+	    gensio_argv_free(so, portv);
 	goto out;
     }
 
     if (startswith(inbuf, "LED", &strtok_data)) {
 	char *name = strtok_r(NULL, ":", &strtok_data);
-	char *str = strtok_r(NULL, "\n", &strtok_data);
-	if (name == NULL) {
+	char *driver = strtok_r(NULL, ":", &strtok_data);
+	char *cfg;
+	const char **argv;
+
+	if (name == NULL || strlen(name) == 0) {
 	    syslog(LOG_ERR, "No LED name given on line %d", lineno);
 	    goto out;
 	}
-	if ((str == NULL) || (strlen(str) == 0)) {
-	    syslog(LOG_ERR, "No LED given on line %d", lineno);
+	if ((driver == NULL) || (strlen(driver) == 0)) {
+	    syslog(LOG_ERR, "No LED driver given on line %d", lineno);
 	    goto out;
 	}
-	handle_led(name, str, lineno);
+	cfg = strtok_r(NULL, "\n", &strtok_data);
+	err = gensio_str_to_argv(so, cfg, NULL, &argv, NULL);
+	if (err) {
+	    syslog(LOG_ERR, "Error parsing LED config: %s\n",
+		   gensio_err_to_str(err));
+	    goto out;
+	}
+	add_led(name, driver, argv, lineno);
+	gensio_argv_free(so, argv);
 	goto out;
     }
 
@@ -972,10 +712,19 @@ handle_config_line(char *inbuf, int len)
     *inbuf = '\0';
     inbuf++;
 
-    timeout = strtok_r(inbuf, ":", &strtok_data);
-    if (timeout == NULL) {
+    timeoutstr = strtok_r(inbuf, ":", &strtok_data);
+    if (timeoutstr == NULL) {
 	syslog(LOG_ERR, "No timeout given on line %d", lineno);
 	goto out;
+    } else {
+	char *end;
+
+	timeout = strtoul(timeoutstr, &end, 0);
+	if (end == timeoutstr || *end != '\0') {
+	    syslog(LOG_ERR, "Invalid timeout '%s' on line %d\n",
+		   timeoutstr, lineno);
+	    goto out;
+	}
     }
 
     devname = strtok_r(NULL, ":", &strtok_data);
@@ -985,57 +734,57 @@ handle_config_line(char *inbuf, int len)
     }
 
     devcfg = strtok_r(NULL, "", &strtok_data);
-    if (devcfg == NULL) {
+    if (devcfg == NULL)
 	/* An empty device config is ok. */
 	devcfg = "";
+
+    err = gensio_str_to_argv(so, devcfg, NULL, &devcfg_argv, NULL);
+    if (err) {
+	syslog(LOG_ERR, "Invalid device config on line %d: %s",
+	       lineno, gensio_err_to_str(err));
+	goto out;
     }
 
-    portconfig(&syslog_eout, portnum, state, timeout, devname, devcfg,
-	       config_num);
+    portconfig(&syslog_eout, portnum, portnum, state, timeout, devname,
+	       devcfg_argv);
 
+    gensio_argv_free(so, devcfg_argv);
  out:
     return 0;
 }
 
-void
+int
 readconfig_init(void)
 {
-    setup_defaults();
+    int err = setup_defaults();
+
+    if (err)
+	return err;
     free_longstrs();
     free_tracefiles();
-#if HAVE_DECL_TIOCSRS485
     free_rs485confs();
-#endif
     free_leds();
 
-    config_num++;
     free_rotators();
+    return 0;
 }
 
 /* Read the specified configuration file and call the routine to
    create the ports. */
 int
-readconfig(char *filename)
+readconfig(FILE *instream)
 {
-    FILE *instream = NULL;
     int linesize = 256;
     char *inbuf = malloc(linesize);
     int  rv = 0, pos = 0;
 
     if (!inbuf) {
 	syslog(LOG_ERR, "Unable to allocate input buffer");
-	return -1;
+	return ENOMEM;
     }
 
     lineno = 0;
 
-    instream = fopen(filename, "r");
-    if (!instream) {
-	syslog(LOG_ERR, "Unable to open config file '%s': %m", filename);
-	rv = -1;
-	goto out_err;
-    }
-
     while (fgets(inbuf + pos, linesize - pos, instream) != NULL) {
 	int len = strlen(inbuf);
 	lineno++;
@@ -1048,7 +797,7 @@ readconfig(char *filename)
 	    new_inbuf = realloc(inbuf, linesize);
 	    if (!new_inbuf) {
 		syslog(LOG_ERR, "Unable to reallocate input buffer");
-		rv = -1;
+		rv = ENOMEM;
 		goto out_err;
 	    }
 	    inbuf = new_inbuf;
@@ -1065,12 +814,7 @@ readconfig(char *filename)
     if (pos > 0)
 	handle_config_line(inbuf, strlen(inbuf));
 
-    /* Delete anything that wasn't in the new config file. */
-    clear_old_port_config(config_num);
-
  out_err:
-    if (instream)
-	fclose(instream);
     free(inbuf);
     return rv;
 }
